/*
 * JCuda - Java bindings for NVIDIA CUDA driver and runtime API
 *
 * Copyright (c) 2009 Marco Hutter - http://www.jcuda.org
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef POINTERUTILS
#define POINTERUTILS

/**
 * The type of a native host memory pointer. The PointerData function
 * will set the MemoryType of the Pointer to the respective value, so
 * that during the corresponding call to releasePointerData the actions
 * appropriate for releasing the respective memory type can be performed
 */
enum MemoryType
{
    NATIVE, POINTERS, DIRECT, ARRAY, ARRAY_COPY
};

/**
 * A structure containing all information necessary for maintaining
 * a pointer to java memory, i.e. to a Java Pointer object. The
 * pointerObject stored in this stucture is only valid for the
 * duration of the JNI call in which it has been initialzed.
 */
typedef struct
{
    /** The global reference to the Java Pointer object */
    jobject pointerObject;

    /** The starting address of the buffer or its array */
    jlong startPointer;

    /** The actual pointer to be used, including offsets */
    jlong pointer;

    /** The type of the memory the pointer points to */
    MemoryType memoryType;

} PointerData;


PointerData* initPointerData(JNIEnv *env, jobject pointerObject);
bool releasePointerData(JNIEnv *env, PointerData* &pointerData, jint mode=0);
void setNativePointer(JNIEnv *env, jobject pointerObject, jlong pointer);
void* getNativePointer(JNIEnv *env, jobject pointerObject);

bool isDirectByteBuffer(JNIEnv *env, jobject object);

int initPointerUtils(JNIEnv *env);

extern jmethodID Buffer_isDirect; // ()Z
extern jmethodID Buffer_hasArray; // ()Z
extern jmethodID Buffer_array; // ()Ljava/lang/Object;

extern jfieldID Pointer_nativePointer; // long
extern jfieldID Pointer_buffer; // Ljava.nio.Buffer;
extern jfieldID Pointer_pointers; // [jcuda.Pointer;
extern jfieldID Pointer_byteOffset; // long

#endif