/*
 * JCuda - Java bindings for NVIDIA CUDA driver and runtime API
 *
 * Copyright (c) 2009 Marco Hutter - http://www.jcuda.org
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

package jcuda.driver;

/**
 * Event creation flags.<br />
 * <br />
 * Most comments are taken from the CUDA reference manual.<br />
 * <br />
 * @see JCudaDriver#cuEventCreate(CUevent, int)
 */
public class CUevent_flags
{
    /**
     * Default event flag
     */
    public static final int CU_EVENT_DEFAULT       = 0;

    /**
     * Event uses blocking synchronization
     */
    public static final int CU_EVENT_BLOCKING_SYNC = 1;

    /**
     * Returns the String identifying the given CUevent_flags
     *
     * @param n The CUevent_flags
     * @return The String identifying the given CUevent_flags
     */
    public static String stringFor(int n)
    {
        switch (n)
        {
            case CU_EVENT_DEFAULT : return "CU_EVENT_DEFAULT";
            case CU_EVENT_BLOCKING_SYNC : return "CU_EVENT_BLOCKING_SYNC";
        }
        return "INVALID CUevent_flags: "+n;
    }

    /**
     * Private constructor to prevent instantiation.
     */
    private CUevent_flags()
    {
    }

}
